﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Windows.Forms;
using HeuristicLab.Core.Views;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.MainForm;
using HeuristicLab.MainForm.WindowsForms;

namespace HeuristicLab.Problems.GeneralizedQuadraticAssignment.Views {
  [View("GQAPAssignmentView")]
  [Content(typeof(GQAPAssignment), IsDefaultView = true)]
  public partial class GQAPAssignmentView : ItemView {
    public new GQAPAssignment Content {
      get { return (GQAPAssignment)base.Content; }
      set { base.Content = value; }
    }

    public GQAPAssignmentView() {
      InitializeComponent();
    }

    #region Register Content Events
    protected override void DeregisterContentEvents() {
      Content.PropertyChanged -= new PropertyChangedEventHandler(Content_PropertyChanged);
      base.DeregisterContentEvents();
    }
    protected override void RegisterContentEvents() {
      base.RegisterContentEvents();
      Content.PropertyChanged += new PropertyChangedEventHandler(Content_PropertyChanged);
    }
    #endregion

    protected override void OnContentChanged() {
      base.OnContentChanged();
      UpdateQuality();
      UpdateAssignment();
    }

    protected override void SetEnabledStateOfControls() {
      base.SetEnabledStateOfControls();
    }

    #region Event Handlers
    private void Content_PropertyChanged(object sender, PropertyChangedEventArgs e) {
      switch (e.PropertyName) {
        case "Quality": UpdateQuality(); break;
        case "FlowDistanceQuality": UpdateFlowDistanceQuality(); break;
        case "InstallationQuality": UpdateInstallationQuality(); break;
        case "OverbookedCapacity": UpdateOverbookedCapacity(); break;
        case "Assignment": UpdateAssignment(); break;
        case "EquipmentNames": UpdateAssignment(); break;
        case "LocationNames": UpdateAssignment(); break;
        default: break;
      }
    }
    #endregion

    private void UpdateQuality() {
      if (InvokeRequired) Invoke((Action)UpdateQuality);
      else {
        if (Content == null) {
          qualityLabel.Text = "-";
        } else {
          qualityLabel.Text = Content.Quality.ToString();
        }
      }
    }

    private void UpdateFlowDistanceQuality() {
      if (InvokeRequired) Invoke((Action)UpdateFlowDistanceQuality);
      else {
        if (Content == null || Content.FlowDistanceQuality == null) {
          flowDistanceQualityLabel.Text = "-";
        } else {
          flowDistanceQualityLabel.Text = Content.FlowDistanceQuality.ToString();
        }
      }
    }

    private void UpdateInstallationQuality() {
      if (InvokeRequired) Invoke((Action)UpdateInstallationQuality);
      else {
        if (Content == null || Content.InstallationQuality == null) {
          installationQualityLabel.Text = "-";
        } else {
          installationQualityLabel.Text = Content.InstallationQuality.ToString();
        }
      }
    }

    private void UpdateOverbookedCapacity() {
      if (InvokeRequired) Invoke((Action)UpdateOverbookedCapacity);
      else {
        if (Content == null || Content.OverbookedCapacity == null) {
          overbookedCapacityLabel.Text = "-";
        } else {
          overbookedCapacityLabel.Text = Content.OverbookedCapacity.ToString();
        }
      }
    }

    private void UpdateAssignment() {
      if (InvokeRequired) Invoke((Action)UpdateAssignment);
      else {
        assignmentTreeView.Nodes.Clear();
        if (Content != null) {
          IntegerVector assignment = Content.Assignment;
          Dictionary<int, TreeNode> locationNodes = new Dictionary<int, TreeNode>();
          for (int i = 0; i < assignment.Length; i++) {
            if (!locationNodes.ContainsKey(assignment[i])) {
              string locationName = assignment[i].ToString();
              if (Content.LocationNames != null && Content.LocationNames.Length > assignment[i])
                locationName = Content.LocationNames[assignment[i]];
              locationNodes.Add(assignment[i], new TreeNode(locationName));
            }
            string equipmentName = i.ToString();
            if (Content.EquipmentNames != null && Content.EquipmentNames.Length > i)
              equipmentName = Content.EquipmentNames[i];
            locationNodes[assignment[i]].Nodes.Add(equipmentName);
          }

          assignmentTreeView.BeginUpdate();
          try {
            foreach (var node in locationNodes.OrderBy(x => x.Key).Select(x => x.Value)) {
              assignmentTreeView.Nodes.Add(node);
              node.Expand();
            }
          } finally { assignmentTreeView.EndUpdate(); }
        }
      }
    }
  }
}
