﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.GeneralizedQuadraticAssignment.Common;

namespace HeuristicLab.Problems.GeneralizedQuadraticAssignment {
  /// <summary>
  /// An operator for analyzing the best solution of Generalized Quadratic Assignment Problems.
  /// </summary>
  [Item("Best GeneralizedQAP Solution Analyzer", "An operator for analyzing the best solution of Generalized Quadratic Assignment Problems.")]
  [StorableClass]
  public sealed class BestGQAPSolutionAnalyzer : SingleSuccessorOperator, IAnalyzer {

    public bool EnabledByDefault {
      get { return true; }
    }

    public LookupParameter<BoolValue> MaximizationParameter {
      get { return (LookupParameter<BoolValue>)Parameters["Maximization"]; }
    }
    public LookupParameter<DoubleMatrix> DistancesParameter {
      get { return (LookupParameter<DoubleMatrix>)Parameters["Distances"]; }
    }
    public LookupParameter<DoubleMatrix> WeightsParameter {
      get { return (LookupParameter<DoubleMatrix>)Parameters["Weights"]; }
    }
    public LookupParameter<DoubleMatrix> InstallationCostsParameter {
      get { return (LookupParameter<DoubleMatrix>)Parameters["InstallationCosts"]; }
    }
    public LookupParameter<DoubleArray> DemandsParameter {
      get { return (LookupParameter<DoubleArray>)Parameters["Demands"]; }
    }
    public LookupParameter<DoubleArray> CapacitiesParameter {
      get { return (LookupParameter<DoubleArray>)Parameters["Capacities"]; }
    }
    public LookupParameter<DoubleValue> TransportationCostsParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["TransportationCosts"]; }
    }
    public LookupParameter<DoubleValue> OverbookedCapacityPenaltyParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["OverbookedCapacityPenalty"]; }
    }
    public ScopeTreeLookupParameter<IntegerVector> AssignmentParameter {
      get { return (ScopeTreeLookupParameter<IntegerVector>)Parameters["Assignment"]; }
    }
    public ScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public ScopeTreeLookupParameter<DoubleValue> FlowDistanceQualityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["FlowDistanceQuality"]; }
    }
    public ScopeTreeLookupParameter<DoubleValue> InstallationQualityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["InstallationQuality"]; }
    }
    public ScopeTreeLookupParameter<DoubleValue> OverbookedCapacityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["OverbookedCapacity"]; }
    }
    public LookupParameter<GQAPAssignment> BestSolutionParameter {
      get { return (LookupParameter<GQAPAssignment>)Parameters["BestSolution"]; }
    }
    public ValueLookupParameter<ResultCollection> ResultsParameter {
      get { return (ValueLookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    public LookupParameter<DoubleValue> BestKnownQualityParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["BestKnownQuality"]; }
    }
    public LookupParameter<IntegerVector> BestKnownSolutionParameter {
      get { return (LookupParameter<IntegerVector>)Parameters["BestKnownSolution"]; }
    }
    public ILookupParameter<StringArray> EquipmentNamesParameter {
      get { return (ILookupParameter<StringArray>)Parameters["EquipmentNames"]; }
    }
    public ILookupParameter<StringArray> LocationNamesParameter {
      get { return (ILookupParameter<StringArray>)Parameters["LocationNames"]; }
    }

    [StorableConstructor]
    private BestGQAPSolutionAnalyzer(bool deserializing) : base(deserializing) { }
    private BestGQAPSolutionAnalyzer(BestGQAPSolutionAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new BestGQAPSolutionAnalyzer(this, cloner);
    }
    public BestGQAPSolutionAnalyzer()
      : base() {
      Parameters.Add(new LookupParameter<BoolValue>("Maximization", "True if the problem is a maximization problem."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Distances", "The distances between the locations."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Weights", "The weights between the equipments."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("InstallationCosts", "The cost of installing equipment x at location y."));
      Parameters.Add(new LookupParameter<DoubleArray>("Demands", "The demands of the equipments."));
      Parameters.Add(new LookupParameter<DoubleArray>("Capacities", "The capacities at the locations."));
      Parameters.Add(new LookupParameter<DoubleValue>("TransportationCosts", "The transportation cost represents the flow-unit per distance-unit cost factor. This value can also be set to 1 if these costs are factored into the weights matrix already."));
      Parameters.Add(new LookupParameter<DoubleValue>("OverbookedCapacityPenalty", "The multiplier for the constraint violation when added to the quality."));
      Parameters.Add(new ScopeTreeLookupParameter<IntegerVector>("Assignment", "The GQAP solutions from which the best solution should be analyzed."));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality", "The qualities of the GQAP solutions which should be analyzed."));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("FlowDistanceQuality", "The flow-distance qualities of the GQAP solutions which should be analyzed."));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("InstallationQuality", "The installation qualities of the GQAP solutions which should be analyzed."));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("OverbookedCapacity", "The overbooked capacities of the GQAP solutions which should be analyzed."));
      Parameters.Add(new LookupParameter<GQAPAssignment>("BestSolution", "The best GQAP solution."));
      Parameters.Add(new ValueLookupParameter<ResultCollection>("Results", "The result collection where the best GQAP solution should be stored."));
      Parameters.Add(new LookupParameter<DoubleValue>("BestKnownQuality", "The quality of the best known solution of this GQAP instance."));
      Parameters.Add(new LookupParameter<IntegerVector>("BestKnownSolution", "The best known solution of this GQAP instance."));
      Parameters.Add(new LookupParameter<StringArray>("EquipmentNames", "A list of names that describes the equipments."));
      Parameters.Add(new LookupParameter<StringArray>("LocationNames", "A list of names that describes the locations."));
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      // BackwardsCompatibility3.3
      #region Backwards compatible code, remove with 3.4
      if (!Parameters.ContainsKey("EquipmentNames"))
        Parameters.Add(new LookupParameter<StringArray>("EquipmentNames", "Optional: A list of names that describes the equipments."));
      if (!Parameters.ContainsKey("LocationNames"))
        Parameters.Add(new LookupParameter<StringArray>("LocationNames", "Optional: A list of names that describes the locations."));
      #endregion
    }

    public override IOperation Apply() {
      var assignments = AssignmentParameter.ActualValue;
      var qualities = QualityParameter.ActualValue;
      var equipmentNames = EquipmentNamesParameter.ActualValue;
      var locationNames = LocationNamesParameter.ActualValue;
      var flowDistanceQualities = FlowDistanceQualityParameter.ActualValue;
      var installationQualities = InstallationQualityParameter.ActualValue;
      var overbookedCapacities = OverbookedCapacityParameter.ActualValue;
      var distances = DistancesParameter.ActualValue;
      var weights = WeightsParameter.ActualValue;
      var installationCosts = InstallationCostsParameter.ActualValue;
      var demands = DemandsParameter.ActualValue;
      var capacities = CapacitiesParameter.ActualValue;
      var transportationCosts = TransportationCostsParameter.ActualValue;
      var overbookedCapacityPenalty = OverbookedCapacityPenaltyParameter.ActualValue;
      var results = ResultsParameter.ActualValue;
      var maximization = MaximizationParameter.ActualValue.Value;
      var bestKnownQuality = BestKnownQualityParameter.ActualValue;

      int bestIndex;
      var tmp = qualities.Select((x, index) => new { Index = index, Value = x.Value });
      if (maximization) bestIndex = tmp.SelectMax(x => x.Value).Index;
      else bestIndex = tmp.SelectMin(x => x.Value).Index;

      if (bestKnownQuality == null || HasSolutionImproved(bestKnownQuality.Value, qualities[bestIndex].Value, maximization)) {
        BestKnownQualityParameter.ActualValue = new DoubleValue(qualities[bestIndex].Value);
        BestKnownSolutionParameter.ActualValue = (IntegerVector)assignments[bestIndex].Clone();
      }

      GQAPAssignment assignment = BestSolutionParameter.ActualValue;
      if (assignment == null) {
        assignment = new GQAPAssignment((IntegerVector)assignments[bestIndex].Clone(), (DoubleValue)qualities[bestIndex].Clone(),
          equipmentNames, locationNames, distances, weights, installationCosts, demands, capacities, transportationCosts,
          overbookedCapacityPenalty, flowDistanceQualities[bestIndex], installationQualities[bestIndex], overbookedCapacities[bestIndex]);
        assignment.Distances = distances;
        BestSolutionParameter.ActualValue = assignment;
        results.Add(new Result("Best GQAP Solution", assignment));
      } else {
        if (HasSolutionImproved(assignment.Quality.Value, qualities[bestIndex].Value, maximization)) {
          assignment.Assignment = (IntegerVector)assignments[bestIndex].Clone();
          assignment.Quality = (DoubleValue)qualities[bestIndex].Clone();
          assignment.EquipmentNames = equipmentNames;
          assignment.LocationNames = locationNames;
          assignment.Distances = distances;
          assignment.Weights = weights;
          assignment.InstallationCosts = installationCosts;
          assignment.Demands = demands;
          assignment.Capacities = capacities;
          assignment.TransportationCosts = transportationCosts;
          assignment.OverbookedCapacityPenalty = overbookedCapacityPenalty;
          assignment.FlowDistanceQuality = (DoubleValue)flowDistanceQualities[bestIndex].Clone();
          assignment.InstallationQuality = (DoubleValue)installationQualities[bestIndex].Clone();
          assignment.OverbookedCapacity = (DoubleValue)overbookedCapacities[bestIndex].Clone();
        }
      }

      return base.Apply();
    }

    private static bool HasSolutionImproved(double oldQuality, double quality, bool maximization) {
      return maximization && oldQuality < quality || !maximization && oldQuality > quality;
    }
  }
}
